#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#if defined(WIN32)
//#  pragma comment(linker, "/subsystem:\"windows\" /entry:\"mainCRTStartup\"")
#  include "glut.h"
#  include "glext.h"
#elif defined(__APPLE__) || defined(MACOSX)
#  include <GLUT/glut.h>
#else
#  define GL_GLEXT_PROTOTYPES
#  include <GL/glut.h>
#endif

#include "normalmap.h"
#include "glsl.h"

/*
** ֥
*/
static GLuint vertShader;
static GLuint fragShader;
static GLuint gl2Program;

/*
** ܥ٥ȥǼ attribute ѿΥϥɥ
*/
GLint tangent;

/*
** 
*/
static const GLfloat lightpos[] = { 4.0, 5.0, 6.0, 1.0 }; /* ֡ */
static const GLfloat lightcol[] = { 1.0, 1.0, 1.0, 1.0 }; /* ľܸ */
static const GLfloat lightamb[] = { 0.1, 0.1, 0.1, 1.0 }; /* Ķ */

/*
** ƥ
*/
#define TEXWIDTH  256                      /* ƥ */
#define TEXHEIGHT 256                      /* ƥι⤵ */

/*
** 
*/
static void init(void)
{
  /* ˡޥåפǼ */
  GLubyte texture[TEXHEIGHT * TEXWIDTH * 4];
  /* ץΥѥ롿󥯷̤ѿ */
  GLint compiled, linked;
  
  /* ƥϥñ̤˵ͤޤƤ */
  glPixelStorei(GL_UNPACK_ALIGNMENT, 4);
  
  /* ƥ硦̾ˡλ */
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
  
  /* ƥη֤ˡλ */
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
  
  /* ƥ˥åȣΥƥĶ */
  glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
  
  /* ˡޥåפκ */
  makeNormalMap(texture, TEXWIDTH, TEXHEIGHT, 20.0, "dotbump.raw");
  
  /* ƥ˥åȣˡޥåפƤ */
  glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, TEXWIDTH, TEXHEIGHT, 0,
    GL_RGBA, GL_UNSIGNED_BYTE, texture);
  
  /*  */
  glClearColor(0.3, 0.3, 1.0, 0.0);
  glEnable(GL_DEPTH_TEST);
  glEnable(GL_CULL_FACE);

  /* ν */
  glEnable(GL_LIGHTING);
  glEnable(GL_LIGHT0);
  glLightfv(GL_LIGHT0, GL_DIFFUSE, lightcol);
  glLightfv(GL_LIGHT0, GL_SPECULAR, lightcol);
  glLightfv(GL_LIGHT0, GL_AMBIENT, lightamb);
  glLightModeli(GL_LIGHT_MODEL_LOCAL_VIEWER, GL_TRUE);
  
  /* GLSL ν */
  if (glslInit()) exit(1);
  
  /* ֥Ȥκ */
  vertShader = glCreateShader(GL_VERTEX_SHADER);
  fragShader = glCreateShader(GL_FRAGMENT_SHADER);
  
  /* Υץɤ߹ */
  if (readShaderSource(vertShader, "bump.vert")) exit(1);
  if (readShaderSource(fragShader, "bump.frag")) exit(1);
  
  /* СƥåΥץΥѥ */
  glCompileShader(vertShader);
  glGetShaderiv(vertShader, GL_COMPILE_STATUS, &compiled);
  printShaderInfoLog(vertShader);
  if (compiled == GL_FALSE) {
    fprintf(stderr, "Compile error in vertex shader.\n");
    exit(1);
  }
  
  /* ե饰ȥΥץΥѥ */
  glCompileShader(fragShader);
  glGetShaderiv(fragShader, GL_COMPILE_STATUS, &compiled);
  printShaderInfoLog(fragShader);
  if (compiled == GL_FALSE) {
    fprintf(stderr, "Compile error in fragment shader.\n");
    exit(1);
  }
  
  /* ץ४֥Ȥκ */
  gl2Program = glCreateProgram();
  
  /* ֥ȤΥץؤϿ */
  glAttachShader(gl2Program, vertShader);
  glAttachShader(gl2Program, fragShader);
  
  /* ֥Ȥκ */
  glDeleteShader(vertShader);
  glDeleteShader(fragShader);
  
  /* ץΥ */
  glLinkProgram(gl2Program);
  glGetProgramiv(gl2Program, GL_LINK_STATUS, &linked);
  printProgramInfoLog(gl2Program);
  if (linked == GL_FALSE) {
    fprintf(stderr, "Link error.\n");
    exit(1);
  }
  
  /* ץŬ */
  glUseProgram(gl2Program);

  /* ƥ˥åȣꤹ */
  glUniform1i(glGetUniformLocation(gl2Program, "texture"), 0);
  
  /* ܥ٥ȥϤ˻Ȥ attribute ѿΥϥɥ */
  tangent = glGetAttribLocation(gl2Program, "tangent");
}

/* ȥåܡѴؿ */
#include "trackball.h"

/* ؿ */
#include "sphere.h"

/*
** 
*/
static void scene(void)
{

  static const GLfloat diffuse[] = { 0.6, 0.1, 0.1, 1.0 };
  static const GLfloat specular[] = { 0.3, 0.3, 0.3, 1.0 };
  
  /*  */
  glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, diffuse);
  glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, specular);
  glMaterialf(GL_FRONT_AND_BACK, GL_SHININESS, 100.0f);
  
  /* ˡޥåפΥޥåԥ󥰳 */
  glEnable(GL_TEXTURE_2D);
  
  /* ȥåܡˤž */
  glMultMatrixd(trackballRotation());
  
  /*  */
  sphere(1.0, 64, 32);
  
  /* ˡޥåפΥޥåԥ󥰽λ */
  glDisable(GL_TEXTURE_2D);
}


/****************************
** GLUT ΥХåؿ **
****************************/

static void display(void)
{
  /* ǥӥ塼Ѵ */
  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();
  
  /* ΰ֤ */
  glLightfv(GL_LIGHT0, GL_POSITION, lightpos);
  
  /* ΰưʪΤ˰ư*/
  glTranslated(0.0, 0.0, -5.0);
  
  /* ̥ꥢ */
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
  
  /*  */
  scene();
  
  /* ֥Хåե */
  glutSwapBuffers();
}

static void resize(int w, int h)
{
  /* ȥåܡ뤹ϰ */
  trackballRegion(w, h);
  
  /* ɥΤӥ塼ݡȤˤ */
  glViewport(0, 0, w, h);
  
  /* ƩѴλ */
  glMatrixMode(GL_PROJECTION);
  
  /* ƩѴν */
  glLoadIdentity();
  gluPerspective(40.0, (double)w / (double)h, 1.0, 100.0);
}

static void idle(void)
{
  /* ̤ؤ */
  glutPostRedisplay();
}

static void mouse(int button, int state, int x, int y)
{
  switch (button) {
  case GLUT_LEFT_BUTTON:
    switch (state) {
    case GLUT_DOWN:
      /* ȥåܡ볫 */
      trackballStart(x, y);
      glutIdleFunc(idle);
      break;
    case GLUT_UP:
      /* ȥåܡ */
      trackballStop(x, y);
      glutIdleFunc(0);
      break;
    default:
      break;
    }
    break;
    default:
      break;
  }
}

static void motion(int x, int y)
{
  /* ȥåܡư */
  trackballMotion(x, y);
}

static void keyboard(unsigned char key, int x, int y)
{
  switch (key) {
  case 'q':
  case 'Q':
  case '\033':
    /* ESC  q  Q 򥿥פ齪λ */
    exit(0);
  default:
    break;
  }
}

/*
** ᥤץ
*/
int main(int argc, char *argv[])
{
  glutInit(&argc, argv);
  glutInitDisplayMode(GLUT_RGBA | GLUT_DEPTH | GLUT_DOUBLE);
  glutCreateWindow(argv[0]);
  glutDisplayFunc(display);
  glutReshapeFunc(resize);
  glutMouseFunc(mouse);
  glutMotionFunc(motion);
  glutKeyboardFunc(keyboard);
  init();
  glutMainLoop();
  return 0;
}
